/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to you under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.shale.usecases.rolodex;

import java.io.IOException;
import java.math.BigDecimal;
import java.text.DecimalFormat;
import java.text.Format;
import java.util.Iterator;
import java.util.Map;

import javax.faces.component.UIComponent;
import javax.faces.component.UIData;
import javax.faces.context.FacesContext;
import javax.faces.context.ResponseWriter;
import javax.faces.render.Renderer;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.shale.util.Messages;

public class WebPagerRenderer extends Renderer {

       
    /**
     * <p>Commons logging utility object static instance.</p>
     */
    private static Log log;
    static {
        log = LogFactory.getLog(WebPagerRenderer.class);
    }

    /**
     * <p>Localized messages for this application.</p>
     */
    private static Messages messages =
      new Messages("org.apache.shale.usecases.view.Bundle");
    
    /**
     * <p>Constant used to create the selected page query parameter.</p>
     */
    private static final String PAGER_SIMPLE_NUMBERLIST = "p";

    /**
     * <p>This basic component doesn't use javascript for 
     * navigation.  This means that the decode logic is in 
     * the encode/rendering phase.</p>
     */
    public void decode(FacesContext context, UIComponent pager) {
        if (log.isTraceEnabled())
            log.trace("decode(FacesContext, UIComponent)");
    }

    /**
     * <p>Locates a parent UIData component, decodes the selected 
     * page query parameter and advances the first row of the 
     * UIData component based on the selected page.</p>
     */
    public void encodeEnd(FacesContext context, UIComponent pager)
        throws IOException {

        if (context == null || pager == null)
            throw new NullPointerException();

        if (!pager.isRendered())
            return;

        UIData data = this.findDataComponent(pager);
        if (data == null)
            throw new NullPointerException(messages.getMessage("webPager.invalid"));

        int startIndex = getPageStartIndex(context, data);
        data.setFirst(Math.max(startIndex, 0));

        ResponseWriter writer = context.getResponseWriter();

        writer.startElement("table",pager);
        writer.writeAttribute("border", "0", "border");

        writer.startElement("tr",pager);
        writer.startElement("td", pager);
        writer.writeAttribute("colspan", "3", "colspan");
        writePageCaption(context, data, pager);
        writer.endElement("td");
        writer.endElement("tr");

        writer.startElement("tr",pager);
        writer.startElement("td", pager);
        writePrevPageLink(context, data, pager);
        writer.endElement("td");
        writer.startElement("td", pager);               
        writePageLinks(context, data, pager);
        writer.endElement("td");
        writer.startElement("td", pager);               
        writeNextPageLink(context, data, pager);
        writer.endElement("td");
        writer.endElement("tr");
        writer.endElement("table");
        writePageIndex(context, data, pager);
        
        writer = null;
        data = null;
    }

    
    /**
     * <p>Writes the previous page link.</p>
     */
    protected void writePrevPageLink(
        FacesContext context,
        UIData data,
        UIComponent pager)
        throws IOException {

        ResponseWriter writer = context.getResponseWriter();
        PageInfo p = prevPage(context, data);

        String styleClass = (String) pager.getAttributes().get("prevPageStyleClass");

        writer.startElement("span", pager);
        
        if (styleClass != null) {
            writer.writeAttribute("class", styleClass, "prevPageStyleClass");
        }
        if (p != null) {

            Object[] params = {new Integer(p.getRowsOnPage())};
            String prevLabel = messages.getMessage("webPager.prevTitle", params);

            StringBuffer url = new StringBuffer(getActionStr(context));
            url.append('?').append(getPageTag(context, data)).append('=');
            url.append(p.getPage());

            writer.startElement("a", pager);
            writer.writeAttribute("href", context.getExternalContext().encodeActionURL(url.toString()), "href");
            writer.writeAttribute("title", prevLabel, "title");
            
            url.setLength(0);
            url.append(getActionStr(context));
            url.setLength(url.indexOf("/", 1));
            String prevImage = (String) pager.getAttributes().get("prevImage");
            
            if (prevImage == null)
               prevImage = messages.getMessage("webPager.prevImage");
            
            url.append(prevImage);
            writer.startElement("img", pager);
            writer.writeAttribute("src", context.getExternalContext().encodeActionURL(url.toString()), "scr");
			writer.endElement("img");
            
            
            writer.endElement("a");
            
            url = null;

        }

        writer.endElement("span");
        p = null;
        writer = null;
        styleClass = null;

    }

    
    /**
     * <p>Renders a hidden input field that keeps track of the current page.</p>
     */
    protected void writePageIndex(FacesContext context, UIData data, UIComponent pager) throws IOException {
        ResponseWriter writer = context.getResponseWriter();
        int page = getPageIndex(context, data);
        writer.startElement("input", pager);
        writer.writeAttribute("name", getPageTag(context, data), "name");
        writer.writeAttribute("type", "hidden", "type");
        writer.writeAttribute("value", Integer.toString(page), "value");
        writer.endElement("input");    
    }
    
    /**
     * <p>Writes a pager caption documenting the total model rows, current page and total pages.</p>
     */
    protected void writePageCaption(FacesContext context, UIData data, UIComponent pager) throws IOException {
        ResponseWriter writer = context.getResponseWriter();
        
        Object[] params = new Object[3];
        params[0] = new Integer(data.getRowCount());
        params[1] = new Integer(getPageIndex(context, data));
        params[2] = new Integer(getPages(data));
        
        String caption = messages.getMessage("webPager.defaultcaption", params);
        
        String styleClass = (String) pager.getAttributes().get("captionStyleClass");

        writer.startElement("span", pager);
        
        if (styleClass != null) {
            writer.writeAttribute("class", styleClass, "pageCaptionStyleClass");
        }
        
        
        writer.writeText(caption, "webPager.defaultcaption");

        writer.endElement("span");
        writer = null;
        styleClass = null;

    }
    
    /**
     * <p>The first ten pages are displayed as individual page links.  
     * The remainder of page links are displayed in increments of ten.</p>
     */
    protected void writePageLinks(FacesContext context, UIData data, UIComponent pager)
        throws IOException {

        ResponseWriter writer = context.getResponseWriter();
        Format fmt = new DecimalFormat("#,##0");

        String styleClass = (String) pager.getAttributes().get("pageLinksStyleClass");

        writer.startElement("span", pager);

        if (styleClass != null) {
            writer.writeAttribute("class", styleClass, "pageLinksStyleClass");
        }
        
        StringBuffer url = new StringBuffer(getActionStr(context));
        url.append('?').append(getPageTag(context, data)).append('=');
        int urlLen = url.length();

        Iterator pi = iterator(context, data);
        while (pi.hasNext()) {

            PageInfo p = (PageInfo) pi.next();

            String ps = fmt.format(new Integer(p.getPage()));
            if (p.isSelected()) {
                writer.write(' ');
                writer.startElement("b", pager);
                writer.writeText(ps, "page");
                writer.endElement("b");
                writer.write(' ');
            } else {

                url.setLength(urlLen);
                url.append(p.getPage());

                writer.write(' ');
                writer.startElement("a", pager);
                writer.writeAttribute("href", context.getExternalContext().encodeActionURL(url.toString()), "href");
                writer.writeAttribute("title", 
                         messages.getMessage("webPager.linkTitle", new Object[] {new Integer(p.getPage())}), "title");
                writer.writeText(ps, "page");
                writer.endElement("a");
                writer.write(' ');
            }
            ps = null;
            p = null;
        }

        writer.endElement("span");

        pi = null;
        url = null;
        fmt = null;
        writer = null;
    }

    /**
     * <p>Generates the next page link.</p>
     */
    protected void writeNextPageLink(FacesContext context, UIData data, UIComponent pager)
        throws IOException {

        ResponseWriter writer = context.getResponseWriter();
        PageInfo p = nextPage(context, data);


        String styleClass = (String) pager.getAttributes().get("nextPageStyleClass");

        writer.startElement("span", pager);
        
        if (styleClass != null) {
            writer.writeAttribute("class", styleClass, "nextPageStyleClass");
        }
        if (p != null) {
            Object[] params = {new Integer(p.getRowsOnPage())};
            String nextLabel = messages.getMessage("webPager.nextTitle", params);
            
            StringBuffer url = new StringBuffer(getActionStr(context));
            url.append('?').append(getPageTag(context, data)).append('=');
            url.append(p.getPage());

            writer.startElement("a", pager);
            writer.writeAttribute("href", context.getExternalContext().encodeActionURL(url.toString()), "href");
            writer.writeAttribute("title", nextLabel, "title");
            
            url.setLength(0);
            url.append(getActionStr(context));
            url.setLength(url.indexOf("/", 1));
            String nextImage = (String) pager.getAttributes().get("nextImage");
            
            if (nextImage == null)
               nextImage = messages.getMessage("webPager.nextImage");
            
            url.append(nextImage);
            writer.startElement("img", pager);
            writer.writeAttribute("src", context.getExternalContext().encodeActionURL(url.toString()), "scr");
            
            
            url = null;

        }

        writer.endElement("span");
        p = null;
        writer = null;
        styleClass = null;

    }

    /**
     * <p>Determines the base URL by calling on the view handler.</p>  
     */
    protected String getActionStr(FacesContext context) {
        String uri = context.getViewRoot().getViewId();
        String url =
            context.getApplication().getViewHandler().getActionURL(
                context,
                uri);

        return url;
    }

    /**
     * <p>Returns the query parameter tag based on the parent UIData 
     * component.</p>
     */
    protected String getPageTag(FacesContext context, UIData data) {
        StringBuffer clientId = new StringBuffer(data.getClientId(context));
        clientId.append(PAGER_SIMPLE_NUMBERLIST);
        return clientId.toString();
    }

    /**
     * <p>Returns the selected page index.  This would be done in the decode method if 
     * this control was not scriptless.</p>
     */
    protected int getPageIndex(FacesContext context, UIData data) {

        int defaultPage = 1;
        int page = 0;
        try {
            Map params = context.getExternalContext().getRequestParameterMap();
            String tag = getPageTag(context, data);
            if (params.containsKey(tag)) {
                page = Integer.parseInt((String) params.get(tag));
            } else {
                page = getCurrentPage(data) + 1;
            }
            tag = null;
            params = null;
        } catch (NumberFormatException e) {
            e.printStackTrace();
        }

        return Math.min(Math.max(page, defaultPage), getPages(data));
    }

    /**
     * <p>Returns the current page based on the UIData component's
     * state.</p>
     */
    protected int getCurrentPage(UIData uidata) {
        if (uidata == null)
            return 0;
        int i = uidata.getRows();
        if (i <= 0)
            return 0;
        else
            return ((int) (uidata.getFirst() / i));
    }
    
    /**
     * <p>Looks for the first data component parent.</p>
     */
    protected UIData findDataComponent(UIComponent uicomponent) {
        return (UIData) findDataParent(uicomponent);
    }

    /**
     * <p>Recursively looks up the component tree for a parent UIData component.</p>
     */
    private UIComponent findDataParent(UIComponent uicomponent) {
        if (uicomponent == null)
            return null;
        if (uicomponent instanceof UIData)
            return uicomponent;
        else
            return findDataParent(uicomponent.getParent());
    }

    /**
     * <p>Returns the rows per page of the UIData component.
     * The value must be five or more.</b> 
     */
    protected int getRowsPerPage(UIData data) {
        int rowsPerPage = Math.max(data.getRows(), 5);
        return rowsPerPage;
    }

    /**
     * <p>Returns the starting row within the UIData's model based
     * on the selected page index, the total rows the model has and 
     * the rows per page.</p> 
     */
    protected int getPageStartIndex(FacesContext context, UIData data) {
        return getPageStartIndex(getPageIndex(context, data), context, data);
    }

    /**
     * <p>Returns the starting row within the UIData's model based
     * on the selected page index, the total rows the model has and 
     * the rows per page.</p> 
     */
    protected int getPageStartIndex(
        int page,
        FacesContext context,
        UIData data) {
        int rowsPerPage = getRowsPerPage(data);
        
        return (page - 1) * rowsPerPage;
    }

    /**
     * <p>Returns the ending row index displayed by the UIData component.</p>
     */
    protected int getPageEndIndex(FacesContext context, UIData data) {
        return getPageEndIndex(getPageIndex(context, data), context, data);
    }

    /**
     * <p>Returns the ending row index displayed by the UIData component.</p>
     */
    protected int getPageEndIndex(
        int page,
        FacesContext context,
        UIData data) {
        int endPageIndex =
            (getPageStartIndex(page, context, data) + getRowsPerPage(data)) - 1;
        return Math.min(endPageIndex, (data.getRowCount() - 1));
    }

    /**
     * <p>Returns the total rows on the selected page.</p> 
     */
    protected int getRowsOnPage(int page, FacesContext context, UIData data) {
        return (
            getPageEndIndex(page, context, data)
                - getPageStartIndex(page, context, data)
                + 1);
    }

    /**
     * <p>Returns the total number of pages based on the size of the model's set
     * and the total rows displayed.</p>
     */
    protected int getPages(UIData data) {
        int pages = 0;
        double rowCount = data.getRowCount();
        double rowsPerPage = getRowsPerPage(data);
        try {
            pages =
                new BigDecimal(rowCount).divide(
                        new BigDecimal(rowsPerPage),
                        0,
                        BigDecimal.ROUND_UP).intValue();
        } catch (Exception e) {
            e.printStackTrace();
        }
        return pages;
    }

    /**
     * <p>Returns PageInfo used to render the previous page links.</p> 
     */
    protected PageInfo prevPage(FacesContext context, UIData data) {
        int pageIndex = getPageIndex(context, data);

        if (pageIndex > 1) {
            int rowsOnPage = getRowsOnPage(pageIndex - 1, context, data);
            return new PageInfo(pageIndex - 1, pageIndex, rowsOnPage);
        }
        return null;
    }

    /**
     * <p>Returns a PageInfo Iterator used to render the individual page
     * links.</p>
     */
    protected Iterator iterator(FacesContext context, UIData data) {
        return new PageInfoIterator(
            getPageIndex(context, data),
            getPages(data));
    }

    /**
     * <p>Returns PageInfo that defines the next link.</p> 
     */
    protected PageInfo nextPage(FacesContext context, UIData data) {
        int pageIndex = getPageIndex(context, data);
        if (data.getRowCount() > data.getRows() && getPages(data) > pageIndex) {
            int rowsOnPage = getRowsOnPage(pageIndex + 1, context, data);
            return new PageInfo(pageIndex + 1, pageIndex, rowsOnPage);
        }
        return null;
    }

    private class PageInfo {

        private int page = 0;
        private boolean selected = false;
        private int rows = 0;

        public PageInfo(int nextIndex, int pageIndex, int rowsOnPage) {
            this(nextIndex, pageIndex);
            rows = rowsOnPage;
        }

        public PageInfo(int nextIndex, int pageIndex) {
            page = nextIndex;
            selected = (pageIndex == nextIndex);
        }

        public int getPage() {
            return page;
        }

        public boolean isSelected() {
            return selected;
        }

        public int getRowsOnPage() {
            return rows;
        }

    }

    private class PageInfoIterator implements Iterator {

        private int pages = 0;
        private int pageIndex = 0;
        private int startWindow = 0;
        private int endWindow = 0;
        private int index = 0;

        public PageInfoIterator(int pageIndex, int pages) {
            this.pages = pages;
            this.pageIndex = Math.min(pageIndex, pages);
            startWindow = (int) ((int) (.1 * (this.pageIndex - 1)) * 10);
            endWindow = Math.min(startWindow + 10, pages);

        }

        public boolean hasNext() {
            return (pages > 1) && (index < pages);
        }

        public Object next() {

            // if the index is within the window, increment by one; otherwise by 10
            if ((index >= startWindow) && (index < endWindow))
                index++;
            else
                index += 10;

            return new WebPagerRenderer.PageInfo(index, pageIndex);
        }

        public void remove() {
        }

    }

    /**
     * <p>Delegates to the super implementation.</p>
     */
    public void encodeBegin(FacesContext context, UIComponent uicomponent)
        throws IOException {
        super.encodeBegin(context, uicomponent);
    }

}
